#define MAT_GROUND 0.
#define MAT_CACTUS 1.
#define MAT_GUY_BODY 2.
#define MAT_GUY_LEGS 3.
#define MAT_GUY_HEAD 4.
#define MAT_GUY_HAT 5.
#define MAT_GUY_KEY 6.

float walkDist() {
    return t * 1.8;
}

#if 0
// inlined for size savings
float sceneSwitch() {
    return smoothstep(0., 1., t)
         - smoothstep(9.9, 10., t)
         + smoothstep(10., 10.1, t)
         - smoothstep(19.9, 20., t)
         + smoothstep(20., 20.1, t)
         - smoothstep(30., 35., t);
}
#endif

vec3 camFunc(inout vec3 ray) {
    float scene = floor(t / 10.);

    if (scene == 0.) {
        // scenery view
        float t2 = clamp(t, 0., 8.);
        pRotate(ray.yz, radians(-2. - 2. * t2));
        pRotate(ray.xz, radians(120. - 8. * t2));
        return vec3(-20, 20, -50);
    } else if (scene == 1.) {
        // face shot of leader from up (and others in background)
#if 0
        vec3 p0 = vec3(-3.3, 7.5, -8.5 - walkDist());
        vec3 target = vec3(0.0, 5.2, - walkDist());
        vec3 dir = p0 - target;
#else
        vec3 target = vec3(0.0, 5.2, - walkDist());
        vec3 dir = vec3(-3.3, 2.3, -13.7);
#endif
        pRotate(dir.xz, radians(-100. + 12. * (t - 10.)));

        ray = lookAt(target + dir, target) * ray;
        return target + dir;
    } /* else { */
        // follow leader from back
        float at = t - 20.;
#if 0
        vec3 p0 = vec3(0, 10., 11. - walkDist());
        vec3 target = vec3(0.0, 5.2 + 0.02 * at, - walkDist());
        vec3 dir = p0 - target;
#else
        vec3 target = vec3(0.0, 5.2 + 0.02 * at, - walkDist());
        vec3 dir = vec3(0, 4.8 - 0.02 * at, 11);
#endif
        ray = lookAt(target + dir, target) * ray;
        return target + 0.4 * (at + 1.) * dir;
    /*} */
#if 0
    // unused scene
    {
        // everyone from side and slightly front
        pRotate(ray.yz, radians(-22.));
        pRotate(ray.xz, radians(76.));
        return vec3(-17.4, 9.9, 4.6 - walkDist());
    }
#endif
}

float walkAngleUpper(float f) {
    if (f <= 0.4)
        return 5. + (f / 0.4) * 85.;
    if (f <= 0.66)
        return 90.;
    if (f <= 0.9)
        return 90. - 85. * (f - 0.66) / (0.9 - 0.66);
    return 5.;
}

float walkAngleLower(float f) {
    if (f <= 0.2)
        return -3.;
    if (f <= 0.4)
        return -3. + (3. + 65.) * (f - 0.2) / (0.4 - 0.2);
    if (f <= 0.7)
        return 65.;
    if (f <= 0.9)
        return 65. - (3. + 65.) * (f - 0.7) / (0.9 - 0.7);
    return -3.;
}

float fArm(vec3 pArm, float sgn, out float dArmConnector, float connectorSmin, float animPhase) {
    vec3 pOrig = pArm;

    float armRadius = 0.15;
    float offsetFromBody = armRadius * 1.15;
    pArm.x -= sgn * offsetFromBody * 0.3; // move arms away from body
    dArmConnector = fCapsule(pArm.yxz, armRadius, offsetFromBody / 2.);
    pArm.x -= sgn * offsetFromBody * 0.7; // move arms away from body

    // 0.0 -> -45
    // 0.5 -> +45
    // 1.0 -> -45
    float o1 = 0.1;
    float o2 = 0.1;
    float upperPartAngle = -45. + 90. * (  smoothstep(0.0 + o2, 0.5 - o1, animPhase)
                                         - smoothstep(0.5 + o1, 1.0 - o2, animPhase));
    float lowerPartAngle = 90.;

    float armPart1Height = 0.4;
    pRotate(pArm.yz, radians(upperPartAngle));  // rotate upper part of arm
    pArm.y += armPart1Height; // move to center point of upper arm capsule
    float dArmPart1 = fCapsule(pArm, armRadius, armPart1Height); // draw it
    //return fOpUnion(length(pOrig) - 0.2, dArmPart1);
    pArm.y += armPart1Height; // move to other endpoint of upper arm capsule

    float armPart2Height = 0.4;
    pRotate(pArm.yz, radians(lowerPartAngle)); // rotate lower part of arm
    pArm.y += armPart2Height; // move to center point of lower arm capsule
    float dArmPart2 = fCapsule(pArm, armRadius, armPart2Height); // draw it
    pArm.y += armPart2Height / 2. + armRadius; // move to other endpoint of lower arm capsule

    return smin(fOpUnion(dArmPart1, dArmPart2), dArmConnector, connectorSmin);
}

float dLeg(vec3 pLeg, float legRadius, float animPhase) {
    float upperPartAngle = 90. - walkAngleUpper(animPhase);
    float lowerPartAngle = walkAngleLower(animPhase);

    float legPart1Height = 0.5;
    pRotate(pLeg.yz, radians(upperPartAngle));  // rotate upper part of leg
    pLeg.y += legPart1Height; // move to center point of upper leg capsule
    float dLegPart1 = fCapsule(pLeg, legRadius, legPart1Height); // draw it
    pLeg.y += legPart1Height; // move to other endpoint of upper leg capsule

    float legPart2Height = 0.5;
    pRotate(pLeg.yz, radians(-upperPartAngle - lowerPartAngle)); // rotate lower part of leg
    pLeg.y += legPart2Height; // move to center point of lower leg capsule
    float dLegPart2 = fCapsule(pLeg, legRadius, legPart2Height); // draw it
    pLeg.y += legPart2Height; // move to other endpoint of lower leg capsule

    float feetX = legRadius;
    float feetY = 0.13;
    float feetZ = 0.7;
    pRotate(pLeg.yz, radians(lowerPartAngle)); // rotate lower part of leg
    pLeg.y += legRadius - feetY / 2.; // move y so bottom of feet == tip of capsule
    pLeg.z += feetZ / 2.;
    float feetRound = 0.08;
    float dFeet = fBox(pLeg, vec3(feetX, feetY, feetZ) - feetRound) - feetRound;

    return smin(fOpUnion(dLegPart1, dLegPart2), dFeet, 0.2);
}

float walkPhase(float f) {
    f *= 5.;
    float ff = fract(f);
    if (ff < 5./6.)
        return floor(f) * 0.2 + 0.2 * ff / (5./6.);
    return floor(f) * 0.2 + 0.2;
}

RaymarchResult dudeSDF(vec3 p) {
    float phase = fract(t * 1.0);
    float dy = 3.57;
    p.y -= dy;

#if 1
    // guard object
    // 1.2 causes artifacts in key
    float guardDist = 1.3;
    float guard = fCylinder(p, guardDist, 5.);
    if (guard > guardDist)
        return makeRR(guard, MAT_GUY_BODY);
#endif

    float bx = 0.7;
    float by = 1.2;
    float bz = 0.3;
    float boxRound = 0.5;
    float dBody = fBox(p, vec3(bx, by, bz) - boxRound) - boxRound;

    // hands
    vec3 pArm = p;
    pArm.x -= bx; // move to point on box where arm will connect
    pArm.y -= by * 0.6; // move to point on box where arm will connect

    float connectorSmin = 0.09;

    float dArmConnector;
    float dArm = fArm(pArm, 1., dArmConnector, connectorSmin, phase);
    dBody = smin(dBody, dArmConnector, connectorSmin);
    dBody = fOpUnion(dBody, dArm);

    pArm.x += 2. * bx; // move to point on box where arm will connect

    dArm = fArm(pArm, -1., dArmConnector, connectorSmin, fract(phase + 0.5));
    dBody = smin(dBody, dArmConnector, connectorSmin);
    dBody = fOpUnion(dBody, dArm);

    RaymarchResult rr = makeRR(dBody, MAT_GUY_BODY);

    // 4Head
    vec3 pHead = p;
    float headRadius = 0.4;
    float headLength = 0.3;
    pHead.y -= by; // move to top of box
    pHead.y -= headLength + headRadius / 1.5;
    float dHead = fCapsule(pHead, headRadius, headLength);
    rr = objUnion(rr, makeRR(dHead, MAT_GUY_HEAD, pHead));

#if 1
    float hatRadius = headRadius + 0.1;
    pHead.y -= headLength * 2. + hatRadius;
    float hatLength = 0.6;
    float dHat = fCapsule(pHead, hatRadius + 0.02 * abs(cos(22. * pHead.y)), hatLength);
    rr = objUnion(rr, makeRR(dHat, MAT_GUY_HAT));
#endif

    // key
    vec3 pKey = p.xzy;
    pRotate(pKey.xz, 2. * t);

    float keyLength = 0.5;
    float keyThick = 0.08;
    // key connecting cylinder
    pKey.y -= keyLength / 2.;
    pKey.y -= bz;
    float dKey = fCylinder(pKey, keyThick, keyLength);

    // key end
    pKey.y -= keyLength;
    pKey = pKey.xzy;
    pKey.x = abs(pKey.x);
    pKey.x -= 0.3;
    float dKeyEnd = fCylinder(pKey, 0.4, keyThick);
    float dHole = fCylinder(pKey, 0.2, keyThick + 0.01);
    dKeyEnd = fOpDifference(dKeyEnd, dHole);

    rr = objUnion(rr, makeRR(fOpUnion(dKey, dKeyEnd), MAT_GUY_KEY));

    // legs
    float legRadius = 0.3;
    float anim = 2. * t;
    float legPart1Angle = 90. * (floor(mod(anim, 2.)) == 1. ? fract(1. - anim) : fract(anim));
    vec3 pLeg = p;
    pLeg.y += by; // move to bottom of the body box
    float legBoxRound = 0.2;
    float legBoxY = 0.3;
    vec3 pLegBox = pLeg;
    pLegBox.z -= 0.01;
    float dLegBox = fBox(pLegBox, vec3(0.77 * 0.8, .21, 0.7 * 0.4) - legBoxRound) - legBoxRound;

    pLeg.y += legRadius; // move to bottom of the body box
    pLeg.x -= bx / 2.; // move to point on box where leg will connect
    float dLeg1 = dLeg(pLeg, legRadius, phase);
    pLeg.x += bx;
    float dLeg2 = dLeg(pLeg, legRadius, fract(phase + 0.5));

    float dLegs = fOpUnion(dLeg1, dLeg2);
    float dLegsAndBox = smin(dLegBox, dLegs, 0.1);
    rr = objUnion(rr, makeRR(dLegsAndBox, MAT_GUY_LEGS));

    return rr;
}

float cactusPart(vec3 p, float radius, float length, float mul) {
    float angle = atan(p.z, p.x);
    return fCapsule(p, radius - 0.04 * abs(sin(mul * angle * PI)), length);
}

#define roadWidth 6.0

RaymarchResult sceneSDF(vec3 p) {
    vec3 pDunes = p;
    pRotate(pDunes.xz, radians(25.));
    bool inRoad = abs(p.x) < roadWidth;
    float groundHeight = inRoad ? 0.0 : 3.8 * sin((abs(p.x) - roadWidth) * 0.1) * sin(p.z * 0.1);

    float dGround = fPlane(p, vec3(0., 1., 0.), groundHeight);
    RaymarchResult rr = makeRR(dGround, MAT_GROUND, p);

    vec3 pCactus1 = p;
    pCactus1.x -= (pCactus1.x < 0.0 ? 1. : -1.) * 5.;
    pCactus1.y += groundHeight;
    vec2 cellIndex = pMod2(pCactus1.xz, vec2(14));

    float cactusProb = 0.7 / abs(cellIndex.x);
    vec3 rng = hashOld33(vec3(cellIndex, 42.));
    vec3 rng2 = hashOld33(vec3(cellIndex, 43.));

    // COMPILE TIME: disable cactus
#if 1
    if (abs(cellIndex.x) > 0.1 && rng.x < cactusProb) {
        // guard object
        VAR float guardDist = 1.1;
        pCactus1.xz += 2.6 * (rng.yz * 2. - 1.);
        float guard = fCylinder(pCactus1, guardDist, 5.);
        if (guard > guardDist) {
            rr = objUnion(rr, makeRR(guard, MAT_CACTUS));
        } else {
            pRotate(pCactus1.xz, radians(202. + 25. * rng2.x));
            float cactusSize = rng2.y;
            vec3 pCactus2 = pCactus1;
            pCactus2.y -= 1.3 + 0.2 * cactusSize;
            pRotate(pCactus2.xy, radians(10. + 35. * rng2.y));
            pCactus2.y -= 1.3 + 0.2 * cactusSize;

            vec3 pCactus3 = pCactus1;
            pRotate(pCactus3.xz, radians(107.));
            pCactus3.y -= 0.8;
            pRotate(pCactus3.xy, radians(12. + 25. * rng2.z));
            pCactus3.y -= 1.5;

            float dCactus1 = cactusPart(pCactus1, 0.3 + 0.1 * cactusSize, 3.0 + cactusSize, 3.);
            float dCactus2 = cactusPart(pCactus2, 0.25, 0.8, 2.);
            float dCactus3 = cactusPart(pCactus3, 0.2, 0.8, 2.);
            float dCactus = fOpUnion(fOpUnion(dCactus1, dCactus2), dCactus3);

            rr = objUnion(rr, makeRR(dCactus, MAT_CACTUS, rng2));
        }
    //} else if (abs(cellIndex.x) >= 1. && abs(cellIndex.x) < 2.) {
    }
#endif

    p.z += floor(walkDist()) + walkPhase(fract(walkDist()));

    // leader dude
    rr = objUnion(rr, dudeSDF(p));

    // patrol
    p.z -= 12.;
    p.x -= 3. * sign(p.x);
    //if (p.x > 0.)
    //    p.x -= 3.;
    //else
    //    p.x += 3.;
    float dudeCell = 5.;
    vec3 q = p;
    if (p.z < dudeCell / 2.) {
        // noop
    } else if (p.z > 3.5 * dudeCell) {
        q.z -= 3.0 * dudeCell;
    } else {
        pMod2(q.zy, vec2(dudeCell,1.e5));
    }
    rr = objUnion(rr, dudeSDF(q));

    return rr;
}

vec3 calcMaterialColor(vec4 materialParam, out vec2 materialSpecularity) {
    VAR color3 col_SAND = vec3(0.400, 0.232, 0.108);
    VAR color3 col_CACTUS = vec3(0.161, 0.380, 0.053);
    VAR color3 col_GUY_BODY = vec3(0.340, 0.000, 0.040);
    VAR color3 col_GUY_HEAD = vec3(0.620, 0.353, 0.353);
    VAR color3 col_GUY_HAT = vec3(0.020, 0.020, 0.020);
    VAR color3 col_GUY_KEY = vec3(0.270, 0.270, 0.270);
    VAR color3 col_GUY_LEGS = vec3(0.006, 0.002, 0.040);

    //VAR float spe1 = 1.0 : 0 _ 1;
    //VAR float spe2 = 1.0 : 0 _ 1;
    materialSpecularity = vec2(0.2, 0.7);
    if (materialParam.w == MAT_GUY_BODY) return col_GUY_BODY;
    if (materialParam.w == MAT_GUY_HAT) return col_GUY_HAT;
    if (materialParam.w == MAT_GUY_KEY) return col_GUY_KEY;
    if (materialParam.w == MAT_GUY_LEGS) return col_GUY_LEGS;

    if (materialParam.w == MAT_GUY_HEAD) {
        //return materialParam.xyz;

        // is front?
        if (materialParam.z < 0.0) {
            vec2 pEye = materialParam.xy;
            pEye.x = abs(pEye.x);
            VAR float eyeX = 0.1387;
            VAR float eyeY = 0.01;
            VAR float eyeD = 0.06;
            float dEye = length(pEye - vec2(eyeX, eyeY));
            if (dEye < eyeD)
                return vec3(0.01);

            vec2 pMouth = materialParam.xy;
            VAR float triMul = 7.2;
            pMouth.y += 0.03 * tri(triMul * pMouth.x);

            VAR float mouthY = 0.2179;
            VAR float mouthDx = 0.1981;
            VAR float mouthDy = 0.02;
            float dMouth = length(pMouth - vec2(eyeX, eyeY));
            if (abs(pMouth.x) < mouthDx && abs(pMouth.y + mouthY) < mouthDy)
                return vec3(0.01);
        }
        return col_GUY_HEAD;
    }

    if (materialParam.w == MAT_GUY_KEY) return col_GUY_KEY;

    materialSpecularity = vec2(0,1);
    if (materialParam.w == MAT_CACTUS) return col_CACTUS * (0.6 + 0.4 * materialParam.y);

    // Ground
    vec3 p = materialParam.xyz;
#define posRng hashOld33(42. + floor(p * 32.))

#define inLaneMarkers (abs(p.x) < 0.1 && mod(p.z, 2.) < 1.) || abs(abs(p.x) - 5.6) < 0.1
    return inLaneMarkers ? vec3(0.9) : abs(p.x) < roadWidth ? vec3(0.03) : col_SAND * (0.7 + 0.3 * posRng.x);
#undef inLaneMarkers
#undef posRng
}
